@@ -2,7 +2,7 @@
 #-# 
 #-# xsnow: let it snow on your desktop
 #-# Copyright (C) 1984,1988,1990,1993-1995,2000-2001 Rick Jansen
-#-# 	      2019,2020,2021 Willem Vermin
+#-# 	      2019,2020,2021,2022,2023,2024 Willem Vermin
 #-# 
 #-# This program is free software: you can redistribute it and/or modify
 #-# it under the terms of the GNU General Public License as published by
@@ -19,12 +19,17 @@
 #-# 
 */
 //
+#include <pthread.h>
+#include <semaphore.h>
 #include <gtk/gtk.h>
 #include <stdlib.h>
 #include <math.h>
 #include <signal.h>
 #include <string.h>
 #include <assert.h>
+
+#include "xsnow-constants.h"
+
 #include "Santa.h"
 #include "birds.h"
 #include "clocks.h"
@@ -37,18 +42,20 @@
 #include "kdtree.h"
 #include "mainstub.h"
 #include "pixmaps.h"
+#include "safe_malloc.h"
 #include "ui.h"
 #include "utils.h"
 #include "windows.h"
 
 
-
 #define NWINGS 8
 #define NBIRDPIXBUFS (3*NWINGS)
 
+#define INACTIVE\
+   (!Flags.ShowBirds || blobals.freeze || !WorkspaceActive())
+
+#define LEAVE_IF_INACTIVE if INACTIVE return TRUE
 
-#define LEAVE_IF_INACTIVE\
-   if (!Flags.ShowBirds || blobals.freeze || !WorkspaceActive()) return TRUE
 
 /* Surface to store current scribbles */
 
@@ -56,21 +63,21 @@ static GdkPixbuf       *bird_pixbufs[NBIRDPIXBUFS];
 static cairo_surface_t *attrsurface = NULL;
 
 static int Nbirds;  // is copied from Flags.Nbirds in init_birds. We cannot have that
-		    //                  // Nbirds is changed outside init_birds
+//                  // Nbirds is changed outside init_birds
 
 
 typedef struct _Birdtype
 {
    float x,y,z;        // position, meters
-		       // x: horizontal
-		       // y: in/out screen
-		       // z: vertical
+   // x: horizontal
+   // y: in/out screen
+   // z: vertical
    float sx,sy,sz;     // velocity, m/sec
    int ix,iy,iz,iw,ih; // pixels
-		       // ix .. in pixels, used to store previous screen parameters
+   // ix .. in pixels, used to store previous screen parameters
    int wingstate, orient;      
    int drawable;       // is in drawable range
-		       // comes in handy at erasing a bird:
+   // comes in handy at erasing a bird:
    int prevx, prevy, prevw, prevh, prevdrawable;
 } BirdType;
 
@@ -88,7 +95,7 @@ static void     show_attr(void);
 static float    attr_maxz(float y);
 static int      do_update_pos_birds(void *); 
 static int      do_wings(void *);
-static int      do_update_speed_birds(void *);
+static void    *do_update_speed_birds(void *);
 static int      do_main_window(void *);
 static void     init_bird_pixbufs(const char *color);
 static void     main_window(void);
@@ -98,6 +105,9 @@ static void     r2i(BirdType *bird);
 static void     i2r(BirdType *bird);
 static int      attrbird_erase(int force);
 static void     init_birds(int start);
+static void     init_bird(BirdType *bird);
+static int      lock(void);
+static int      unlock(void);
 
 
 static float time_update_pos_birds     = 0.01;
@@ -109,22 +119,23 @@ static struct kdtree *kd = NULL;
 static BirdType *birds = NULL;
 static BirdType attrbird;
 
+static sem_t sem;
 
 void birds_ui()
 {
-   UIDO(ShowBirds         , birds_erase(1); attrbird_erase(1); );
-   UIDO(BirdsOnly         , ClearScreen();                     );
-   UIDO(Neighbours        ,                                    );
-   UIDO(Anarchy           ,                                    );
-   UIDO(PrefDistance      ,                                    );
-   UIDO(ViewingDistance   , attrbird2surface();                );
-   UIDO(BirdsSpeed        , birds_set_speed();                 );
-   UIDO(AttrFactor        ,                                    );
-   UIDO(DisWeight         ,                                    );
-   UIDO(FollowWeight      ,                                    );
-   UIDO(BirdsScale        , birds_set_scale();                 );
-   UIDO(ShowAttrPoint     , attrbird_erase(1);                 );
-   UIDO(AttrSpace         , show_attr();                     );
+   UIDO(ShowBirds         , lock(); birds_erase(1); unlock(); attrbird_erase(1);     );
+   UIDO(BirdsOnly         , ClearScreen();  do_change_attr(NULL);  );
+   UIDO(Neighbours        ,                                        );
+   UIDO(Anarchy           ,                                        );
+   UIDO(PrefDistance      ,                                        );
+   UIDO(ViewingDistance   , attrbird2surface();                    );
+   UIDO(BirdsSpeed        , birds_set_speed();                     );
+   UIDO(AttrFactor        ,                                        );
+   UIDO(DisWeight         ,                                        );
+   UIDO(FollowWeight      ,                                        );
+   UIDO(BirdsScale        , birds_set_scale();                     );
+   UIDO(ShowAttrPoint     , attrbird_erase(1);                     );
+   UIDO(AttrSpace         , show_attr();                           );
    UIDOS(BirdsColor       , 
 	 birds_init_color(); 
 	 ClearScreen(););
@@ -148,6 +159,21 @@ void birds_ui()
    }
 }
 
+void birds_sem_init()
+{
+   sem_init(&sem,0,1);
+}
+
+int lock()
+{
+   return sem_wait(&sem);
+}
+
+int unlock()
+{
+   return sem_post(&sem);
+}
+
 static void normalize_speed(BirdType *bird, float speed)
 {
    float v2 = sq3(bird->sx, bird->sy, bird->sz);
@@ -159,7 +185,6 @@ static void normalize_speed(BirdType *bird, float speed)
    bird->sz *= a;
 }
 
-
 static float scale(float y)
 {
    float s;
@@ -259,149 +284,157 @@ void birds_set_scale()
    attrbird2surface();
 }
 
-int do_update_speed_birds(void *d)
+void *do_update_speed_birds(void *d)
 {
-   if (Flags.Done)
-      return FALSE;
-   LEAVE_IF_INACTIVE;
-   P("do_update_speed_birds %d\n",counter);
-
-   kd_free(kd);
-   kd = kd_create(3);
-
-   int i;
-
-   for (i=0; i<Nbirds; i++)
+   while(1)
    {
-      BirdType *bird = &birds[i];
-      kd_insert3f(kd, bird->x, bird->y, bird->z, bird);
+      if (!(Flags.Done || INACTIVE))
+      {
+
+	 lock();
+
+	 P("do_update_speed_birds %d\n",global.counter++);
+
+	 kd_free(kd);
+	 kd = kd_create(3);
+
+	 int i;
+
+	 for (i=0; i<Nbirds; i++)
+	 {
+	    BirdType *bird = &birds[i];
+	    kd_insert3f(kd, bird->x, bird->y, bird->z, bird);
+	 }
+
+	 int sumnum         = 0;
+	 float summeandist  = 0;
+	 for (i=0; i<Nbirds; i++)
+	 {
+	    if (drand48() < Flags.Anarchy*0.01)
+	       continue;
+	    BirdType *bird = &birds[i];
+
+	    struct kdres *result = kd_nearest_range3f(kd,bird->x,bird->y,bird->z,blobals.range);
+	    float sumsx    = 0;
+	    float sumsy    = 0;
+	    float sumsz    = 0;
+	    float sumprefx = 0;
+	    float sumprefy = 0;
+	    float sumprefz = 0;
+	    float sumdist  = 0;
+	    int   num      = 0;
+	    while (!kd_res_end(result))
+	    {
+	       float x,y,z;
+	       BirdType *b = (BirdType *)kd_res_item3f(result, &x, &y, &z);
+	       kd_res_next(result);
+	       if (bird == b)
+		  continue;
+	       num++;
+
+	       // sum the speeds of neighbour birds:
+
+	       sumsx += b->sx;
+	       sumsy += b->sy;
+	       sumsz += b->sz;
+
+	       float dist = sqrtf((bird->x-x)*(bird->x-x)+
+		     (bird->y-y)*(bird->y-y)+(bird->z-z)*(bird->z-z));
+
+	       float prefx=0, prefy=0, prefz=0;
+	       P("prefxyz %f\n",dist);
+	       if (dist > 1e-6)
+		  prefxyz(bird, dist, Flags.PrefDistance, x, y, z, &prefx, &prefy, &prefz);
+	       sumprefx += prefx;
+	       sumprefy += prefy;
+	       sumprefz += prefz;
+	       sumdist  += dist;
+
+	    }
+	    kd_res_free(result);
+
+	    // meanprefx,y,z: mean optimal coordinates with respect to other birds
+	    float meanprefx, meanprefy, meanprefz, meandist;
+	    P("num: %d\n",num);
+	    if (num > 0)
+	    {
+	       meanprefx     = sumprefx / num;
+	       meanprefy     = sumprefy / num;
+	       meanprefz     = sumprefz / num;
+	       meandist      = sumdist  / num;
+	       summeandist  += meandist;
+	       P("prefx - x ... %f %f %f %f\n",meanprefx - bird->x, meanprefy - bird->y, meanprefz - bird->z, meandist);
+	    }
+	    sumnum +=num;
+	    // adjust speed to other birds, p is weight for own speed
+	    if (num > 0)
+	    {
+	       int p = (100-Flags.FollowWeight)*0.1;
+	       bird->sx = (sumsx + p*bird->sx)/(p+1+num);
+	       bird->sy = (sumsy + p*bird->sy)/(p+1+num);
+	       bird->sz = (sumsz + p*bird->sz)/(p+1+num);
+	    }
+	    // adjust speed to obtain desired distance to other birds
+	    if (num > 0)
+	    {
+	       float q = Flags.DisWeight*0.4;
+	       bird->sx += q*(meanprefx - bird->x);
+	       bird->sy += q*(meanprefy - bird->y);
+	       bird->sz += q*(meanprefz - bird->z);
+	       P("%d %f %f %f, %f %f %f\n",i,meanprefx,meanprefy,meanprefz,bird->x,bird->y,bird->z);
+	    }
+
+
+	    // attraction of center:
+
+	    float dx = attrbird.x - bird->x;
+	    float dy = attrbird.y - bird->y;
+	    float dz = attrbird.z - bird->z;
+
+	    float f = Flags.AttrFactor*0.01f*0.05f;
+
+	    bird->sx += f*dx;
+	    bird->sy += f*dy;
+	    bird->sz += f*dz;
+
+	    // limit vertical speed
+
+	    const float phs = 0.8;
+	    float hs = sqrtf(sq2(bird->sx, bird->sy));
+	    if (fabs(bird->sz) > phs*hs)
+	       bird->sz = fsignf(bird->sz)*phs*hs;
+
+	    // randomize:
+	    {
+	       const float p  = 0.4;  //  0<=p<=1 the higher the more random
+	       bird->sx += bird->sx*p*drand48();
+	       bird->sy += bird->sy*p*drand48();
+	       bird->sz += bird->sz*p*drand48();
+	    }
+
+	    normalize_speed(bird,blobals.meanspeed*(0.9+drand48()*0.2));
+	 }
+	 float meannum = (float)sumnum/(float)Nbirds;
+	 blobals.mean_distance = summeandist/Nbirds;
+	 P("meannum %f %f\n",meannum,blobals.range);
+
+	 if (meannum < Flags.Neighbours)
+	 {
+	    if (blobals.range < 0.1)
+	       blobals.range = 0.1;
+	    if (meannum < Nbirds-1)
+	       blobals.range *=1.1;
+	    if (blobals.range > blobals.maxrange)
+	       blobals.range /=1.1;
+	 }
+	 else
+	    blobals.range /=1.1;
+
+	 unlock();
+      }
+      usleep((useconds_t)(time_update_speed_birds*1.0e6));
    }
-
-   int sumnum         = 0;
-   float summeandist  = 0;
-   for (i=0; i<Nbirds; i++)
-   {
-      if (drand48() < Flags.Anarchy*0.01)
-	 continue;
-      BirdType *bird = &birds[i];
-
-      struct kdres *result = kd_nearest_range3f(kd,bird->x,bird->y,bird->z,blobals.range);
-      float sumsx    = 0;
-      float sumsy    = 0;
-      float sumsz    = 0;
-      float sumprefx = 0;
-      float sumprefy = 0;
-      float sumprefz = 0;
-      float sumdist  = 0;
-      int   num      = 0;
-      while (!kd_res_end(result))
-      {
-	 float x,y,z;
-	 BirdType *b = (BirdType *)kd_res_item3f(result, &x, &y, &z);
-	 kd_res_next(result);
-	 if (bird == b)
-	    continue;
-	 num++;
-
-	 // sum the speeds of neighbour birds:
-
-	 sumsx += b->sx;
-	 sumsy += b->sy;
-	 sumsz += b->sz;
-
-	 float dist = sqrtf((bird->x-x)*(bird->x-x)+
-	       (bird->y-y)*(bird->y-y)+(bird->z-z)*(bird->z-z));
-
-	 float prefx=0, prefy=0, prefz=0;
-	 P("prefxyz %f\n",dist);
-	 if (dist > 1e-6)
-	    prefxyz(bird, dist, Flags.PrefDistance, x, y, z, &prefx, &prefy, &prefz);
-	 sumprefx += prefx;
-	 sumprefy += prefy;
-	 sumprefz += prefz;
-	 sumdist  += dist;
-
-      }
-      kd_res_free(result);
-
-      // meanprefx,y,z: mean optimal coordinates with respect to other birds
-      float meanprefx, meanprefy, meanprefz, meandist;
-      P("num: %d\n",num);
-      if (num > 0)
-      {
-	 meanprefx     = sumprefx / num;
-	 meanprefy     = sumprefy / num;
-	 meanprefz     = sumprefz / num;
-	 meandist      = sumdist  / num;
-	 summeandist  += meandist;
-	 P("prefx - x ... %f %f %f %f\n",meanprefx - bird->x, meanprefy - bird->y, meanprefz - bird->z, meandist);
-      }
-      sumnum +=num;
-      // adjust speed to other birds, p is weight for own speed
-      if (num > 0)
-      {
-	 int p = (100-Flags.FollowWeight)*0.1;
-	 bird->sx = (sumsx + p*bird->sx)/(p+1+num);
-	 bird->sy = (sumsy + p*bird->sy)/(p+1+num);
-	 bird->sz = (sumsz + p*bird->sz)/(p+1+num);
-      }
-      // adjust speed to obtain desired distance to other birds
-      if (num > 0)
-      {
-	 float q = Flags.DisWeight*0.4;
-	 bird->sx += q*(meanprefx - bird->x);
-	 bird->sy += q*(meanprefy - bird->y);
-	 bird->sz += q*(meanprefz - bird->z);
-	 P("%d %f %f %f, %f %f %f\n",i,meanprefx,meanprefy,meanprefz,bird->x,bird->y,bird->z);
-      }
-
-
-      // attraction of center:
-
-      float dx = attrbird.x - bird->x;
-      float dy = attrbird.y - bird->y;
-      float dz = attrbird.z - bird->z;
-
-      float f = Flags.AttrFactor*0.01f*0.05f;
-
-      bird->sx += f*dx;
-      bird->sy += f*dy;
-      bird->sz += f*dz;
-
-      // limit vertical speed
-
-      const float phs = 0.8;
-      float hs = sqrtf(sq2(bird->sx, bird->sy));
-      if (fabs(bird->sz) > phs*hs)
-	 bird->sz = fsignf(bird->sz)*phs*hs;
-
-      // randomize:
-      {
-	 const float p  = 0.4;  //  0<=p<=1 the higher the more random
-	 bird->sx += bird->sx*p*drand48();
-	 bird->sy += bird->sy*p*drand48();
-	 bird->sz += bird->sz*p*drand48();
-      }
-
-      normalize_speed(bird,blobals.meanspeed*(0.9+drand48()*0.2));
-   }
-   float meannum = (float)sumnum/(float)Nbirds;
-   blobals.mean_distance = summeandist/Nbirds;
-   P("meannum %f %f\n",meannum,blobals.range);
-
-   if (meannum < Flags.Neighbours)
-   {
-      if (blobals.range < 0.1)
-	 blobals.range = 0.1;
-      if (meannum < Nbirds-1)
-	 blobals.range *=1.1;
-      if (blobals.range > blobals.maxrange)
-	 blobals.range /=1.1;
-   }
-   else
-      blobals.range /=1.1;
-
-   return TRUE;
+   return NULL;
    (void)d;
 }
 
@@ -410,6 +443,9 @@ int do_update_pos_birds(void *d)
    if (Flags.Done)
       return FALSE;
    LEAVE_IF_INACTIVE;
+
+   lock();
+
    P("do_update_pos_birds %d %d\n",Nbirds,counter++);
    double dt;
    dt = time_update_pos_birds;
@@ -425,6 +461,9 @@ int do_update_pos_birds(void *d)
       bird->y += dt*bird->sy;
       bird->z += dt*bird->sz;
    }
+
+   unlock();
+
    return TRUE;
    (void)d;
 }
@@ -438,6 +477,8 @@ int birds_draw(cairo_t *cr)
    int before;
    int i;
 
+   lock();
+
    for (before=0; before<2; before++)
    {
       if(before && Flags.FollowSanta && !Flags.BirdsOnly)
@@ -626,15 +667,20 @@ int birds_draw(cairo_t *cr)
 	 }
       }    // i-loop
    }  // before-loop
+
+   unlock();
+
    return TRUE;
 }
 
 int birds_erase(int force)
 {
+   // locking by caller
    if(global.IsDouble)
       return TRUE;
    if(!force)
       LEAVE_IF_INACTIVE;
+
    P("birds_erase %d\n",counter++);
    int i;
    for (i=0; i<Nbirds; i++)
@@ -679,12 +725,15 @@ int attrbird_erase(int force)
 
 void init_birds(int start)
 {
+   lock();
+
    int i;
    if(!global.IsDouble)
       birds_erase(1);
    P("nbirds: %d %d\n",start,Flags.Nbirds);
    // Bbirds+1 to prevent allocating zero bytes:
    birds = (BirdType *)realloc(birds,sizeof(BirdType)*(Flags.Nbirds+1));
+   REALLOC_CHECK(birds);
    if (kd)
       kd_free(kd);
    kd = kd_create(3);
@@ -726,6 +775,8 @@ void init_birds(int start)
 
       kd_insert3f(kd, bird->x, bird->y, bird->z, NULL);
    }
+
+   unlock();
 }
 
 
@@ -734,6 +785,9 @@ static int do_wings(void *d)
    if (Flags.Done)
       return FALSE;
    LEAVE_IF_INACTIVE;
+
+   lock();
+
    int i;
    for (i=0; i<Nbirds; i++)
    {
@@ -742,6 +796,9 @@ static int do_wings(void *d)
       if (bird->wingstate >= NWINGS)
 	 bird->wingstate = 0;
    }
+
+   unlock();
+
    return TRUE;
    (void)d;
 }
@@ -785,7 +842,8 @@ void birds_set_speed()
 int do_main_window(void *d)
 {
    (void) d;
-   if (blobals.maxix != global.SnowWinWidth || blobals.maxiz != global.SnowWinHeight)
+   P("do_main_window: %d %d %d %d\n",blobals.maxix,global.SnowWinWidth,blobals.maxiz,global.SnowWinHeight);
+   if (blobals.maxix != (int)global.SnowWinWidth || blobals.maxiz != (int)global.SnowWinHeight)
    {
       P("do_main_window\n");
       main_window();
@@ -800,7 +858,7 @@ static void main_window()
    blobals.maxiz = global.SnowWinHeight;
    blobals.maxiy = (blobals.maxix+blobals.maxiz)/2;
 
-   P("%d %d %d\n",blobals.maxix,blobals.maxiy,blobals.maxiz);
+   P("in main_window: %d %d %d\n",blobals.maxix,blobals.maxiy,blobals.maxiz);
 
    blobals.maxz = blobals.maxx*(float)blobals.maxiz/(float)blobals.maxix;
    blobals.maxy = blobals.maxx*(float)blobals.maxiy/(float)blobals.maxix;
@@ -839,13 +897,37 @@ static void init_bird_pixbufs(const char *color)
    }
 }
 
+void init_bird(BirdType *bird)
+{
+   bird->x = 0;
+   bird->y = 0;
+   bird->z = 0;
+   bird->sx = 0;
+   bird->sy = 0;
+   bird->sz = 0;
+   bird->ix = 0;
+   bird->iy = 0;
+   bird->iz = 0;
+   bird->iw = 0;
+   bird->ih = 0;
+   bird->wingstate = 0;
+   bird->orient    = 0;
+   bird->drawable  = 0;
+   bird->prevx     = 0;
+   bird->prevy     = 0;
+   bird->prevw     = 0;
+   bird->prevh     = 0;
+   bird->prevdrawable = 0;
+}
+
 float attr_maxz(float y)
 {
    // convert screen coordinate iz to fractional birds coordinate z
    // Create a dummy bird:
    BirdType bird;
+   init_bird(&bird);                              // just to keep cppcheck happy
    bird.ix = 100;                                 // does not matter what
-   bird.y  = y; //0.5*blobals.maxy;                    // distance in the centre of maxy
+   bird.y  = y;                                   // distance in the centre of maxy
    bird.iz = blobals.maxiz*0.01*Flags.AttrSpace;  // the desired maximum vertical screen value
    i2r(&bird);
    return bird.z;
@@ -870,27 +952,44 @@ int do_change_attr(void *d)
 {
    (void)d;
    // move attraction point in the range
-   // x: 0.3 .. 0.7
+   // x: 0.1 .. 0.8
    // y: 0.4 .. 0.6
    // z: dependent on Flags.AttrSpace
    if (Flags.Done)
       return FALSE;
    if (Flags.FollowSanta && !Flags.BirdsOnly)
       return TRUE;
-   float p = drand48()*attr_maxz(attrbird.y)/blobals.maxz;
-   P("change attr %f\n",p);
+   //
+   // determine random x, y and z values for the attraction point
+   // for x and y: determine screen coordinates and convert
+   // these to bird coordinates.
+   // Note that birds_set_attraction_point_relative() expects
+   // fractional coordinates
+   //
+
+
+   float y = 0.4+drand48()*0.2;
+
+   attrbird.ix = (0.1+drand48()*0.8)*blobals.maxix;
+   attrbird.y  = y*blobals.maxy;
+   attrbird.iz = 120;  // of no importance
+
+   float z = drand48()*attr_maxz(attrbird.y)/blobals.maxz;
+
+   i2r(&attrbird);
+
+   float x = attrbird.x/blobals.maxx;
+   P("attrbird.x: %f\n",attrbird.x/blobals.maxx);
+
    attrbird_erase(1);
-   birds_set_attraction_point_relative(
-	 0.3+drand48()*0.4, 
-	 0.4+drand48()*0.2, 
-	 p
-	 );
+   birds_set_attraction_point_relative(x, y, z);
    r2i(&attrbird);
+   P("change attr %f %f %f , %d %d\n",x,y,z,attrbird.ix,attrbird.iz);
    attrbird2surface();
    return TRUE;
 }
 
-void birds_init ()
+void birds_init()
 {
    init_bird_pixbufs("black"); // just to have pixbufs we can throw away
    birds_init_color();
@@ -915,14 +1014,16 @@ void birds_init ()
 
       clear_flags();
       add_to_mainloop(PRIORITY_HIGH,time_update_pos_birds,     do_update_pos_birds   );
-      add_to_mainloop(PRIORITY_HIGH,time_update_speed_birds,   do_update_speed_birds );
       add_to_mainloop(PRIORITY_HIGH,time_wings,                do_wings              );
       add_to_mainloop(PRIORITY_DEFAULT,time_change_attr,       do_change_attr        );
       add_to_mainloop(PRIORITY_DEFAULT,time_main_window,       do_main_window        );
+
+      static pthread_t thread;
+      P("birds speed thread start\n");
+      pthread_create(&thread,NULL,do_update_speed_birds,NULL);
       main_window();
    }
 
-
    attrbird.x = blobals.maxx/2;
    attrbird.y = blobals.maxy/2;
    attrbird.z = blobals.maxz/2;
@@ -941,6 +1042,8 @@ void birds_init ()
 
    birds_set_speed();
    init_birds(0);
+   P("Calling do_change_attr():\n");
+   do_change_attr(NULL);
 
    attrbird2surface();
 
