@@ -2,7 +2,7 @@
 #-# 
 #-# xsnow: let it snow on your desktop
 #-# Copyright (C) 1984,1988,1990,1993-1995,2000-2001 Rick Jansen
-#-# 	      2019,2020,2021 Willem Vermin
+#-# 	      2019,2020,2021,2022,2023,2024 Willem Vermin
 #-# 
 #-# This program is free software: you can redistribute it and/or modify
 #-# it under the terms of the GNU General Public License as published by
@@ -42,20 +42,27 @@
  * Reals dealing with time are declared as double. 
  * Other reals as float
  */
+#include <pthread.h>
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
 #include <X11/Intrinsic.h>
 #include <X11/extensions/Xdbe.h>
+#include <X11/extensions/Xinerama.h>
 #include <ctype.h>
 #include <gtk/gtk.h>
 #include <cairo-xlib.h>
+#include <gsl/gsl_version.h>
 #include <stdlib.h>
 #include <math.h>
 #include <signal.h>
 #include <stdio.h>
 #include <string.h>
+#include "mygettext.h"
 
+#include "xsnow-constants.h"
+
+#include "aurora.h"
 #include "birds.h"
 #include "clocks.h"
 #include "docs.h"
@@ -63,7 +70,7 @@
 #include "fallensnow.h"
 #include "flags.h"
 #include "mainstub.h"
-#include "meteo.h"
+#include "meteor.h"
 #include "moon.h"
 #include "Santa.h"
 #include "scenery.h"
@@ -82,6 +89,7 @@
 #include "treesnow.h"
 #include "loadmeasure.h"
 #include "selfrep.h"
+#include "xdo.h"
 
 #include "vroot.h"
 
@@ -116,10 +124,14 @@ static int Argc;
 static int          DoRestart      = 0;
 static guint        draw_all_id    = 0;
 static guint        drawit_id      = 0;
-static int          Xorig;
-static int          Yorig;
 static GtkWidget   *TransA         = NULL;
 static char        *SnowWinName    = NULL;
+static int          wantx          = 0;
+static int          wanty          = 0;
+static int          IsSticky       = 0;
+static int          X11cairo       = 0;
+static int          PrevW          = 0;
+static int          PrevH          = 0;
 
 /* Colo(u)rs */
 static const char *BlackColor  = "black";
@@ -134,13 +146,13 @@ static void   drawit(cairo_t *cr);
 static void   restart_do_draw_all(void);
 static void   set_below_above(void);
 static void   DoAllWorkspaces(void);
-static void   X11WindowById(Window *xwin, char **xwinname);
+static void   X11WindowById(Window *xwin);
 static int    HandleX11Cairo(void);
 static int    StartWindow(void);
 static void   SetWindowScale(void);
-static void   movewindow(void);
 static void   GetDesktopSession(void);
-
+static void   rectangle_draw(cairo_t *cr);
+static void   mybindtestdomain(void);
 
 // callbacks
 static int do_displaychanged(void *);
@@ -168,8 +180,9 @@ static gboolean     on_draw_event(GtkWidget *widget, cairo_t *cr, gpointer user_
  *   - scenery (trees etc)
  *   - stars
  *   - moon + halo
- *   - meteorites
+ *   - meteors
  *   - birds
+ *   - aurora
  *
  * Selection of window to draw in
  * ==============================
@@ -247,10 +260,16 @@ int main_c(int argc, char *argv[])
    signal(SIGINT,  SigHandler);
    signal(SIGTERM, SigHandler);
    signal(SIGHUP,  SigHandler);
-   srand48((long int)(wallcl()*1.0e6));
+   P("srand %d\n",(int)(fmod(wallcl()*1.0e6,1.0e8)));
+   srand48((int)(fmod(wallcl()*1.0e6,1.0e8)));
 
+   memset(&global,0,sizeof(global));
 
-   memset((void *)&global,0,sizeof(global));
+   XInitThreads();
+
+   fallen_sem_init();
+   aurora_sem_init();
+   birds_sem_init();
 
    global.counter            = 0;
    global.cpufactor          = 1.0;
@@ -262,9 +281,15 @@ int main_c(int argc, char *argv[])
    global.FluffCount         = 0;
 
    global.SnowWin            = 0;
+   global.WindowOffsetX      = 0;
+   global.WindowOffsetY      = 0;
    global.DesktopSession     = NULL;
    global.CWorkSpace         = 0;
+   global.ChosenWorkSpace    = 0;
+   global.NVisWorkSpaces     = 1;
+   global.VisWorkSpaces[0]   = 0;
    global.WindowsChanged     = 0;
+   global.ForceRestart       = 0;
 
    global.FsnowFirst         = NULL;
    global.MaxScrSnowDepth    = 0;
@@ -278,23 +303,18 @@ int main_c(int argc, char *argv[])
 
    global.Wind               = 0;
    global.Direction          = 0;
-   global.WindMax            = 100.0;
-   global.NewWind = 0;
+   global.WindMax            = 500.0;
+   global.NewWind            = 100.0;
 
    global.HaltedByInterrupt  = 0;
    global.Message[0]         = 0;
 
    global.SantaPlowRegion    = 0;
 
+   global.DoCapella          = DOCAPELLA;
+
    int i;
-   // make a copy of all flags, before gtk_init() maybe removes some.
-   // we need this at a restart of the program.
 
-   Argv = (char**) malloc((argc+1)*sizeof(char**));
-   for (i=0; i<argc; i++)
-      Argv[i] = strdup(argv[i]);
-   Argv[argc] = NULL;
-   Argc = argc;
 
    InitFlags();
    // we search for flags that only produce output to stdout,
@@ -331,13 +351,64 @@ int main_c(int argc, char *argv[])
       }
 #endif
    }
+   int rc = HandleFlags(argc, argv);
+
+   handle_language(0); // the langues used is from flags or .xsnowrc
+		       //                     this changes env LANGUAGE accordingly
+		       //                     so that the desired translation is in effect
+
+   mybindtestdomain();
+   switch(rc)
+   {
+      case -1:   // wrong flag
+		 //PrintVersion();
+	 Thanks();
+	 return 1;
+	 break;
+      case 1:    // manpage or help
+		 //         Ok, this cannot happen, is already caught above
+	 return 0;
+	 break;
+      default:  // ditto
+		//PrintVersion();
+	 break;
+   }
+   PrintVersion();
+
+   printf(_("Available languages are: %s\n"),LANGUAGES);
+   /* Eskimo warning */
+   if (strstr(Flags.SnowColor,"yellow") != NULL)
+      printf("\n%s\n\n",_("Warning: don't eat yellow snow!"));
+
+   // make a copy of all flags, before gtk_init() maybe removes some.
+   // we need this at a restart of the program.
+   // we need to remove 
+   //   -screen n
+   //   -lang c
+
+   Argv = (char**) malloc((argc+1)*sizeof(char**));
+   Argc = 0;
+   for (i=0; i<argc; i++)
+   {
+      if ((strcmp("-screen",argv[i])==0) || (strcmp("-lang",argv[i])==0))
+	 i++; // found -screen or -lang
+      else 
+	 Argv[Argc++] = strdup(argv[i]);
+   }
+   Argv[Argc] = NULL;
    GetDesktopSession();
    if(!strncasecmp(global.DesktopSession,"bspwm",5))
    {
-      printf("For optimal resuts, add to your bspwmrc:\n");
+      printf(_("For optimal resuts, add to your %s:\n"),"bspwmrc");
       printf("   bspc rule -a Xsnow state=floating border=off\n");
    }
-   printf("Xsnow running in GTK version: %s\n",ui_gtk_version());
+   printf("GTK version: %s\n",ui_gtk_version());
+   printf("GSL version: ");
+#ifdef GSL_VERSION
+   printf("%s\n",GSL_VERSION);
+#else
+   printf(_("unknown\n"));
+#endif
 
    // Circumvent wayland problems:before starting gtk: make sure that the 
    // gdk-x11 backend is used.
@@ -345,29 +416,13 @@ int main_c(int argc, char *argv[])
    setenv("GDK_BACKEND","x11",1);
    if (getenv("WAYLAND_DISPLAY")&&getenv("WAYLAND_DISPLAY")[0])
    {
-      printf("Detected Wayland desktop\n");
+      printf(_("Detected Wayland desktop\n"));
       global.IsWayland = 1;
    }
    else
       global.IsWayland = 0;
 
    gtk_init(&argc, &argv);
-   int rc = HandleFlags(argc, argv);
-   switch(rc)
-   {
-      case -1:   // wrong flag
-	 PrintVersion();
-	 Thanks();
-	 return 1;
-	 break;
-      case 1:    // manpage or help
-	 //         Ok, this cannot happen, is already caught above
-	 return 0;
-	 break;
-      default:  // ditto
-	 PrintVersion();
-	 break;
-   }
    if (!Flags.NoConfig)
       WriteFlags();
 
@@ -376,7 +431,7 @@ int main_c(int argc, char *argv[])
 
    if(Flags.CheckGtk && !ui_checkgtk() && !Flags.NoMenu)
    {
-      printf("Xsnow needs gtk version >= %s, found version %s\n",ui_gtk_required(),ui_gtk_version());
+      printf(_("Xsnow needs gtk version >= %s, found version %s \n"),ui_gtk_required(),ui_gtk_version());
 
       if(!ui_run_nomenu())
       {
@@ -386,14 +441,23 @@ int main_c(int argc, char *argv[])
       else
       {
 	 Flags.NoMenu = 1;
-	 printf("Continuing with flag '-nomenu'\n");
+	 printf("%s %s",_("Continuing with flag"),"'-nomenu'\n");
       }
    }
 
    global.display = XOpenDisplay(Flags.DisplayName);
+   global.xdo = xdo_new_with_opened_display(global.display,NULL,0);
+   if(global.xdo == NULL)
+   {
+      I("xdo problems\n");
+      exit(1);
+   }
+   global.xdo->debug = 0;
+
    XSynchronize(global.display,dosync);
    XSetErrorHandler(XsnowErrors);
    int screen = DefaultScreen(global.display);
+   global.Screen = screen;
    global.Black = BlackPixel(global.display, screen);
    global.White = WhitePixel(global.display, screen);
 
@@ -403,6 +467,8 @@ int main_c(int argc, char *argv[])
 	 Flags.TreeColor = strdup(DefaultFlags.TreeColor); 
       if (!ValidColor(Flags.SnowColor))
 	 Flags.SnowColor = strdup(DefaultFlags.SnowColor); 
+      if (!ValidColor(Flags.SnowColor2))
+	 Flags.SnowColor2 = strdup(DefaultFlags.SnowColor2); 
       if (!ValidColor(Flags.BirdsColor))
 	 Flags.BirdsColor = strdup(DefaultFlags.BirdsColor); 
       WriteFlags();
@@ -413,8 +479,8 @@ int main_c(int argc, char *argv[])
 
    if (global.display == NULL) {
       if (Flags.DisplayName == NULL) Flags.DisplayName = getenv("DISPLAY");
-      (void) fprintf(stderr, "%s: cannot connect to X server %s\n", argv[0],
-	    Flags.DisplayName ? Flags.DisplayName : "(default)");
+      fprintf(stderr, _("%s: cannot connect to X server %s\n"), argv[0],
+	    Flags.DisplayName ? Flags.DisplayName : _("(default)"));
       exit(1);
    }
    //
@@ -458,20 +524,50 @@ int main_c(int argc, char *argv[])
       return 1;
    }
 
+
+#define DOIT_I(x,d,v) OldFlags.x = Flags.x;
+#define DOIT_L(x,d,v) DOIT_I(x,d,v);
+#define DOIT_S(x,d,v) OldFlags.x = strdup(Flags.x);
+   DOITALL;
+#include "undefall.inc"
+
+
+   OldFlags.FullScreen = !Flags.FullScreen;
+
+   BlackPix = AllocNamedColor(BlackColor, global.Black);
+
+   if(global.Desktop)
+   {
+      XSelectInput(global.display, global.Rootwindow, StructureNotifyMask| SubstructureNotifyMask);
+   }
+   else
+      XSelectInput(global.display, global.SnowWin, StructureNotifyMask|SubstructureNotifyMask);
+
+   ClearScreen();   // without this, no snow, scenery etc. in KDE; this seems to be a vintage comment
+
+
+   if(!Flags.NoMenu && !global.XscreensaverMode)
+   {
+      ui();
+      ui_gray_erase(global.Trans);
+      ui_set_sticky(Flags.AllWorkspaces);
+      add_to_mainloop(PRIORITY_DEFAULT, time_desktop_type, do_show_desktop_type);
+   }
    Flags.Done = 0;
    windows_init();
    moon_init();
+   aurora_init();
    Santa_init();
    birds_init();
    scenery_init();
    snow_init();
-   meteo_init();
+   meteor_init();
    wind_init();
    stars_init();
-   fallensnow_init();
    blowoff_init();
    treesnow_init();
    loadmeasure_init();
+   fallensnow_init();
 
    add_to_mainloop(PRIORITY_DEFAULT, time_displaychanged,     do_displaychanged     );
    add_to_mainloop(PRIORITY_DEFAULT, time_event,              do_event              );
@@ -485,39 +581,14 @@ int main_c(int argc, char *argv[])
 
    HandleCpuFactor();
 
-#define DOIT_I(x,d,v) OldFlags.x = Flags.x;
-#define DOIT_L(x,d,v) DOIT_I(x,d,v);
-#define DOIT_S(x,d,v) OldFlags.x = strdup(Flags.x);
-   DOITALL;
-#include "undefall.inc"
+   fflush(stdout);
 
+   DoAllWorkspaces(); // to set global.ChosenWorkSpace
 
-   OldFlags.FullScreen = !Flags.FullScreen;
-
-   BlackPix = AllocNamedColor(BlackColor, global.Black);
-
-   // events
-   if(global.Desktop)
-   {
-      XSelectInput(global.display, global.Rootwindow, StructureNotifyMask| SubstructureNotifyMask);
-   }
-   else
-      XSelectInput(global.display, global.SnowWin, StructureNotifyMask|SubstructureNotifyMask);
-
-   ClearScreen();   // without this, no snow, scenery etc. in KDE; this seems to be a vintage comment
-
-   if(!Flags.NoMenu && !global.XscreensaverMode)
-   {
-      ui();
-      ui_gray_erase(global.Trans);
-      ui_set_sticky(Flags.AllWorkspaces);
-      add_to_mainloop(PRIORITY_DEFAULT, time_desktop_type, do_show_desktop_type);
-   }
-
+   P("Entering main loop\n");
    // main loop
    gtk_main();
 
-
    if (SnowWinName) free(SnowWinName);
 
    XClearWindow(global.display, global.SnowWin);
@@ -527,8 +598,13 @@ int main_c(int argc, char *argv[])
    if (DoRestart)
    {
       sleep(0);
-      printf("Xsnow restarting: %s\n",Argv[0]);
+      printf("Xsnow %s: ",_("restarting"));
+      int k = 0;
+      while (Argv[k])
+	 printf("%s ",Argv[k++]);
+      printf("\n");
       fflush(NULL);
+      setenv("XSNOW_RESTART","yes",1);
       execvp(Argv[0],Argv);
    }
    else
@@ -557,7 +633,7 @@ void set_below_above()
    }
 }
 
-void X11WindowById(Window *xwin, char **xwinname)
+void X11WindowById(Window *xwin)
 {
    *xwin = 0;
    // user supplied window id:
@@ -569,9 +645,11 @@ void X11WindowById(Window *xwin, char **xwinname)
    if (Flags.XWinInfoHandling)
    {
       // user ask to point to a window
-      printf("Click on a window ...\n");
-      *xwin = XWinInfo(xwinname);
-      if (*xwin == 0)
+      printf(_("Click on a window ...\n"));
+      fflush(stdout);
+      int rc;
+      rc = xdo_select_window_with_click(global.xdo,xwin);
+      if (rc == XDO_ERROR)
       {
 	 fprintf(stderr,"XWinInfo failed\n");
 	 exit(1);
@@ -601,10 +679,10 @@ void GetDesktopSession()
 	    break;
       }
       if (desktopsession)
-	 printf("Detected desktop session: %s\n",desktopsession);
+	 printf(_("Detected desktop session: %s\n"),desktopsession);
       else
       {
-	 printf("Could not determine desktop session\n");
+	 printf(_("Could not determine desktop session\n"));
 	 desktopsession = (char *)"unknown_desktop_session";
       }
 
@@ -614,7 +692,7 @@ void GetDesktopSession()
 
 int StartWindow()
 {
-   int X11cairo     = 0;
+   P("Entering StartWindow...\n");
 
    global.Trans     = 0;
    global.xxposures = 0;
@@ -626,7 +704,7 @@ int StartWindow()
    global.Rootwindow = DefaultRootWindow(global.display);
    Window xwin;
    // see if user chooses window
-   X11WindowById(&xwin, NULL);
+   X11WindowById(&xwin);
    if (xwin)
    {
       P("StartWindow xwin%#lx\n",xwin);
@@ -636,7 +714,7 @@ int StartWindow()
    else if(Flags.ForceRoot)
    {
       // user specified to run on root window
-      printf("Trying to snow in root window\n");
+      printf(_("Trying to snow in root window\n"));
       global.SnowWin = global.Rootwindow;
       if (getenv("XSCREENSAVER_WINDOW"))
       {
@@ -657,13 +735,16 @@ int StartWindow()
 
 
       GdkWindow *gdkwin; (void)gdkwin;
-      int rc = make_trans_window(gtkwin,
-	    1,                   // full screen 
+      int rc = make_trans_window(global.display, gtkwin,
+	    Flags.Screen,        // full screen or xinerama
 	    Flags.AllWorkspaces, // sticky 
 	    Flags.BelowAll,      // below
 	    1,                   // dock
 	    NULL,                // gdk_window
-	    &xwin                // x11_window
+	    &xwin,               // x11_window
+	    &wantx,              // make_trans_window tries to place the window here,
+				 //                      but, depending on window manager that does not always succeed
+	    &wanty
 	    );
       if (rc)
       {
@@ -677,11 +758,12 @@ int StartWindow()
 	 P("calling set_below_above\n");
 	 set_below_above();
 	 global.SnowWin = xwin;
-	 printf("Using transparent window\n");
+	 printf(_("Using transparent window\n"));
+	 P("wantx, wanty: %d %d\n",wantx,wanty);
 	 if(!strncasecmp(global.DesktopSession,"fvwm",4) ||
 	       !strncasecmp(global.DesktopSession,"lxqt",4))
 	 {
-	    printf("The transparent snow-window is probably not click-through, alas..\n");
+	    printf(_("The transparent snow-window is probably not click-through, alas..\n"));
 	 }
 
       }
@@ -690,28 +772,36 @@ int StartWindow()
 	 global.Desktop  = 1;
 	 X11cairo = 1;
 	 // use rootwindow, pcmanfm or Desktop:
-	 printf("Cannot create transparent window\n");
+	 printf(_("Cannot create transparent window\n"));
 	 if (!strcasecmp(global.DesktopSession,"enlightenment"))
-	    printf("NOTE: xsnow will probably run, but some glitches are to be expected.\n");
+	    printf(_("NOTE: xsnow will probably run, but some glitches are to be expected.\n"));
 	 if(!strcasecmp(global.DesktopSession,"twm"))
 	    printf("NOTE: you probably need to tweak 'Lift snow on windows' in the 'settings' panel.\n");
 	 // if envvar DESKTOP_SESSION == LXDE, search for window with name pcmanfm
+	 // largest_window_with_name uses name argument as regex, hence the '^' and '$'.
+	 // Note that the name is still case-insensitive
 	 if (!strncmp(global.DesktopSession,"LXDE",4) && 
-	       (xwin = FindWindowWithName(global.display,"pcmanfm")))
+	       (xwin = largest_window_with_name(global.xdo,"^pcmanfm$"))
+	    )
 	 {
-	    printf("LXDE session found, using window 'pcmanfm'.\n");
+	    printf(_("LXDE session found, using window 'pcmanfm'.\n"));
 	    P("lxdefound: %d %#lx\n",lxdefound,*xwin);
 	 }
-	 else if ((xwin = FindWindowWithName(global.display,"Desktop")))
+	 else if (
+	       (xwin = largest_window_with_name(global.xdo,"^Desktop$"))
+	       )
 	 {
-	    printf("Using window 'Desktop'.\n");
+	    printf(_("Using window 'Desktop'.\n"));
 	 }
 	 else
 	 {
-	    printf("Using root window\n");
+	    printf(_("Using root window\n"));
 	    xwin = global.Rootwindow;
 	 }
 	 global.SnowWin = xwin;
+	 int winw,winh;
+	 if (Flags.Screen >=0 && global.Desktop)
+	    xinerama(global.display,Flags.Screen,&wantx,&wanty,&winw,&winh);
       }
    }
 
@@ -719,6 +809,14 @@ int StartWindow()
    {
       HandleX11Cairo();
       drawit_id = add_to_mainloop1(PRIORITY_HIGH, time_draw_all, do_drawit, CairoDC);
+      global.WindowOffsetX = 0;
+      global.WindowOffsetY = 0;
+   }
+   else
+   {
+      // see also windows.c
+      global.WindowOffsetX = wantx;
+      global.WindowOffsetY = wanty;
    }
 
    global.IsDouble = global.Trans || global.UseDouble; // just to be sure ...
@@ -729,27 +827,43 @@ int StartWindow()
    else
       SnowWinName = strdup("no name");
    XFree(x.value);
+
+   if(!X11cairo)
+   {
+      xdo_move_window(global.xdo,global.SnowWin,wantx,wanty);
+      P("wantx wanty: %d %d\n",wantx,wanty);
+   }
+   xdo_wait_for_window_map_state(global.xdo,global.SnowWin,IsViewable);
    InitDisplayDimensions();
-   printf("Snowing in %#lx: %s %d+%d %dx%d\n",global.SnowWin,SnowWinName,global.SnowWinX,global.SnowWinY,global.SnowWinWidth,global.SnowWinHeight);
 
-   Xorig = global.SnowWinX;
-   Yorig = global.SnowWinY;
+   global.SnowWinX = wantx;
+   global.SnowWinY = wanty;
 
-   movewindow();
+   printf(_("Snowing in %#lx: %s %d+%d %dx%d\n"),global.SnowWin,SnowWinName,global.SnowWinX,global.SnowWinY,global.SnowWinWidth,global.SnowWinHeight);
+   PrevW = global.SnowWinWidth;
+   PrevH = global.SnowWinHeight;
+
+   P("woffx: %d %d\n",global.WindowOffsetX,global.WindowOffsetY);
+
+   fflush(stdout);
 
    SetWindowScale();
    if (global.XscreensaverMode && !Flags.BlackBackground)
       SetBackground();
 
+   /*
+    * to remove titlebar in mwm. Alas, this results in no visuals at all in xfce.
+    XSetWindowAttributes attr;
+    attr.override_redirect = True;
+    XChangeWindowAttributes(global.display,global.SnowWin,CWOverrideRedirect,&attr);
+    */
    return TRUE;
 }
 
 int HandleX11Cairo()
 {
-   XWindowAttributes attr;
-   XGetWindowAttributes(global.display,global.SnowWin,&attr);
-   int w = attr.width;
-   int h = attr.height;
+   unsigned int w,h;
+   xdo_get_window_size(global.xdo, global.SnowWin, &w, &h);
    Visual *visual = DefaultVisual(global.display,DefaultScreen(global.display));
    int rcv;
    P("double: %d\n",Flags.UseDouble);
@@ -770,18 +884,20 @@ int HandleX11Cairo()
       CairoSurface = cairo_xlib_surface_create(global.display, backBuf, visual, w, h);
       global.UseDouble = 1;
       global.IsDouble  = 1;
-      printf("Using double buffer: %#lx. %dx%d\n",backBuf,w,h);
+      printf(_("Using double buffer: %#lx. %dx%d\n"),backBuf,w,h);
       rcv = TRUE;
    }
 #endif
    if(!dodouble)
    {
       CairoSurface = cairo_xlib_surface_create(global.display, global.SnowWin, visual, w, h);
-      printf("NOT using double buffering:");
+      printf(_("NOT using double buffering:"));
       if (Flags.UseDouble)
-	 printf(" because double buffering is not available on this system\n");
+	 printf(_(" because double buffering is not available on this system\n"));
       else
-	 printf(" on your request.\n");
+	 printf(_(" on your request.\n"));
+      printf(_("NOTE: expect some flicker.\n"));
+      fflush(stdout);
       rcv = FALSE;
    }
 
@@ -790,6 +906,24 @@ int HandleX11Cairo()
 
    CairoDC = cairo_create(CairoSurface);
    cairo_xlib_surface_set_size(CairoSurface,w,h);
+   global.SnowWinWidth  = w;
+   global.SnowWinHeight = h;
+   if (Flags.Screen >= 0 && global.Desktop)
+   {
+      int winx, winy, winw, winh;
+      int rc = xinerama(global.display,Flags.Screen,&winx, &winy, &winw, &winh);
+      //if (rc && winx==0 && winy==0)
+      if(rc)
+      {
+	 global.SnowWinX      = winx;
+	 global.SnowWinY      = winy;
+	 global.SnowWinWidth  = winw;
+	 global.SnowWinHeight = winh;
+      }
+      cairo_rectangle(CairoDC,global.SnowWinX,global.SnowWinY,global.SnowWinWidth,global.SnowWinHeight);
+      P("clipsnow %d %d %d %d\n",global.SnowWinX,global.SnowWinY,global.SnowWinWidth,global.SnowWinHeight);
+      cairo_clip(CairoDC);
+   }
    return rcv;
 }
 
@@ -798,20 +932,37 @@ void DoAllWorkspaces()
    if(Flags.AllWorkspaces)
    {
       P("stick\n");
-      if (global.Trans)
-      {
-	 gtk_window_stick(GTK_WINDOW(TransA));
-      }
+      set_sticky(1);
    }
    else
    {
       P("unstick\n");
-      if (global.Trans)
+      set_sticky(0);
+      if (Flags.Screen >=0)
+	 global.ChosenWorkSpace = global.VisWorkSpaces[Flags.Screen];
+      else
+	 global.ChosenWorkSpace = global.VisWorkSpaces[0];
+   }
+   ui_set_sticky(Flags.AllWorkspaces);
+}
+
+int set_sticky(int s)
+{
+   int r = IsSticky;
+   if (global.Trans)
+   {
+      if(s)
       {
+	 IsSticky = 1;
+	 gtk_window_stick(GTK_WINDOW(TransA));
+      }
+      else
+      {
+	 IsSticky = 0;
 	 gtk_window_unstick(GTK_WINDOW(TransA));
       }
    }
-   ui_set_sticky(Flags.AllWorkspaces);
+   return r;
 }
 
 // here we are handling the buttons in ui
@@ -831,29 +982,35 @@ int do_ui_check(void *d)
    scenery_ui();
    birds_ui();
    snow_ui();
-   meteo_ui();
+   meteor_ui();
    wind_ui();
    stars_ui();
    fallensnow_ui();
    blowoff_ui();
    treesnow_ui();
    moon_ui();
+   aurora_ui();
    ui_ui();
 
-   UIDO (CpuLoad             , HandleCpuFactor();               );
-   UIDO (Transparency        ,                                  );
-   UIDO (Scale               ,                                  );
-   UIDO (OffsetS             , DisplayDimensions();             );
-   UIDO (OffsetY             , UpdateFallenSnowRegions();       );
-   UIDO (NoFluffy            , ClearScreen();                   );
-   UIDO (AllWorkspaces       , DoAllWorkspaces();               );
-   UIDO (BelowAll            , set_below_above();               );
-   UIDOS(BackgroundFile      ,                                  );
-   UIDO (BlackBackground     ,                                  );
+   UIDO (CpuLoad             , HandleCpuFactor();                 );
+   UIDO (Transparency        ,                                    );
+   UIDO (Scale               ,                                    );
+   UIDO (OffsetS             , DisplayDimensions();               );
+   UIDO (OffsetY             , 
+	 int x = global.DoCapella;
+	 global.DoCapella=0; 
+	 UpdateFallenSnowRegionsWithLock(); 
+	 global.DoCapella = x;                                    );
+   UIDO (NoFluffy            , ClearScreen();                     );
+   UIDO (AllWorkspaces       , DoAllWorkspaces();                 );
+   UIDO (BelowAll            , set_below_above();                 );
+   UIDOS(BackgroundFile      ,                                    );
+   UIDO (BlackBackground     ,                                    );
 
    if (Flags.Changes > 0)
    {
       WriteFlags();
+      set_buttons();
       P("-----------Changes: %d\n",Flags.Changes);
    }
    Flags.Changes = 0;
@@ -862,15 +1019,6 @@ int do_ui_check(void *d)
 }
 
 
-void movewindow()
-{
-   if(!global.Desktop)
-      return;
-   XMoveWindow(global.display,global.SnowWin,0,0);
-   //InitDisplayDimensions();
-   P("movewindow: Orig: %d %d %d %d\n",Xorig,Yorig,global.SnowWinWidth,global.SnowWinHeight);
-}
-
 int do_displaychanged(void *d)
 {
    // if we are snowing in the desktop, we check if the size has changed,
@@ -885,22 +1033,30 @@ int do_displaychanged(void *d)
 
    if (!global.Desktop)
       return TRUE;
+
+   if (global.ForceRestart)
+   {
+      DoRestart = 1;
+      Flags.Done = 1;
+      I(_("Restart due to change of screen or language settings ...\n"));
+   }
+   else
    {
       unsigned int w,h;
       Display* display = XOpenDisplay(Flags.DisplayName);
       Screen* screen   = DefaultScreenOfDisplay(display);
       w = WidthOfScreen(screen);
       h = HeightOfScreen(screen);
-      P("width height: %d %d\n",w,h);
+      P("width height: %d %d %d %d\n",w,h,global.Wroot,global.Hroot);
       if(global.Wroot != w || global.Hroot != h)
       {
 	 DoRestart = 1;
 	 Flags.Done = 1;
-	 printf("Restart due to change of display settings...\n");
+	 I(_("Restart due to change of display settings...\n"));
       }
       XCloseDisplay(display);
-      return TRUE;
    }
+   return TRUE;
 }
 
 int do_event(void *d)
@@ -949,6 +1105,7 @@ int do_event(void *d)
 void RestartDisplay()
 {
    P("Restartdisplay: %d W: %d H: %d\n",global.counter++,global.SnowWinWidth,global.SnowWinHeight);
+   fflush(stdout);
    InitFallenSnow();
    init_stars();
    EraseTrees();
@@ -986,13 +1143,13 @@ int do_show_desktop_type(void *d)
       return TRUE;
    char *s;
    if (global.IsWayland)
-      s = (char *)"Wayland (Expect some slugginess)";
+      s = (char *)_("Wayland (Expect some slugginess)");
    else if (global.IsCompiz)
       s = (char *)"Compiz";
    else
-      s = (char *)"Probably X11";
+      s = (char *)_("Probably X11");
    char t[128];
-   snprintf(t,64,"%s. Snow window: %#lx",s,global.SnowWin);
+   snprintf(t,64,_("%s. Snow window: %#lx"),s,global.SnowWin);
    ui_show_desktop_type(t);
    return TRUE;
    (void)d;
@@ -1004,18 +1161,29 @@ int do_testing(void *d)
    (void)d;
    if (Flags.Done)
       return FALSE;
-
    return TRUE;
+   int xret,yret;
+   unsigned int wret,hret;
+   xdo_get_window_location(global.xdo, global.SnowWin,&xret,&yret,NULL);
+   xdo_get_window_size(global.xdo, global.SnowWin, &wret, &hret);
+   P("%d wxh %d %d %d %d    %d %d %d %d \n",global.counter++,global.SnowWinX,global.SnowWinY,global.SnowWinWidth,global.SnowWinHeight,xret,yret,wret,hret);
+   P("WorkspaceActive, chosen: %d %ld\n",WorkspaceActive(),global.ChosenWorkSpace);
+   P("vis:");
+   int i;
+   for (i=0; i<global.NVisWorkSpaces; i++)
+      printf(" %ld",global.VisWorkSpaces[i]);
+   printf("\n");
+   printf("offsets: %d %d\n",global.WindowOffsetX,global.WindowOffsetY);
    global.counter++;
-   XWindowAttributes attr;
-   XGetWindowAttributes(global.display,global.SnowWin,&attr);
 
-   P("%d wxh %d %d %d %d    %d %d %d %d \n",global.counter,global.SnowWinX,global.SnowWinY,global.SnowWinWidth,global.SnowWinHeight,attr.x,attr.y,attr.width,attr.height);
+   for (i=0; i<global.NVisWorkSpaces; i++)
+      printf("%d: visible: %d %ld\n",global.counter,i,global.VisWorkSpaces[i]);
+
+   P("current workspace: %ld\n",global.CWorkSpace);
 
    return TRUE;
 }
 
-
 void SigHandler(int signum)
 {
    global.HaltedByInterrupt = signum;
@@ -1033,7 +1201,7 @@ int XsnowErrors(Display *dpy, XErrorEvent *err)
 
    if (++count > countmax)
    {
-      snprintf(global.Message,sizeof(global.Message),"More than %d errors, I quit!",countmax);
+      snprintf(global.Message,sizeof(global.Message),_("More than %d errors, I quit!"),countmax);
       Flags.Done = 1;
    }
    return 0;
@@ -1053,14 +1221,27 @@ gboolean on_draw_event(GtkWidget *widget, cairo_t *cr, gpointer user_data)
 
 int do_drawit(void *cr)
 {
-   P("do_drawit %d %p\n",counter++,cr);
+   P("do_drawit %d %p\n",global.counter++,cr);
    drawit((cairo_t*)cr);
    return TRUE;
 }
 
+
 void drawit(cairo_t *cr)
 {
-   P("drawit %d %p\n",counter++,(void *)cr);
+   static int counter = 0;
+   P("drawit %d %p\n",global.counter++,(void *)cr);
+   // Due to instabilities at the start of xsnow: spurious detection of
+   //   user intervention; resizing of screen; etc.
+   //   placement of scenery, stars etc is repeated a few times at the
+   //   start of xsnow.
+   // This is not harmful, but a bit annoying, so we do not draw anything 
+   //   the first few times this function is called.
+   if (counter*time_draw_all < 1.5)
+   {
+      counter++;
+      return;
+   }
 
    if (Flags.Done)
       return;
@@ -1080,6 +1261,7 @@ void drawit(cairo_t *cr)
       stars_erase(); // not really necessary
       birds_erase(0);
       snow_erase(1);
+      aurora_erase();
       XFlush(global.display);
    }
 
@@ -1093,15 +1275,27 @@ void drawit(cairo_t *cr)
       cairo_fill(cr);
       */
 
+   cairo_save(cr);
+   int tx = 0;
+   int ty = 0;
+   if (X11cairo)
+   {
+      tx = global.SnowWinX;
+      ty = global.SnowWinY;
+   }
+   P("tx ty: %d %d\n",tx,ty);
+   cairo_translate(cr,tx,ty);
+
    int skipit = Flags.BirdsOnly || !WorkspaceActive();
 
    if (!skipit)
    {
       stars_draw(cr);
-      meteo_draw(cr);
 
       P("moon\n");
       moon_draw(cr);
+      aurora_draw(cr);
+      meteor_draw(cr);
       scenery_draw(cr);
    }
 
@@ -1109,7 +1303,7 @@ void drawit(cairo_t *cr)
    birds_draw(cr);
 
    if (skipit)
-      return;
+      goto end;;
 
    fallensnow_draw(cr);
 
@@ -1120,6 +1314,12 @@ void drawit(cairo_t *cr)
 
    snow_draw(cr);
 
+end:
+   if (Flags.Outline)
+      rectangle_draw(cr);
+
+   cairo_restore(cr);
+
    XFlush(global.display);
 }
 
@@ -1144,23 +1344,21 @@ int do_display_dimensions(void *d)
    if (!SnowWinChanged)
       return TRUE;
    SnowWinChanged = 0;
-   static int prevw = 0, prevh = 0;
-   P("%d do_display_dimensions %dx%d %dx%d\n",global.counter++,global.SnowWinWidth,global.SnowWinHeight,prevw,prevh);
+
+   P("%d do_display_dimensions %dx%d %dx%d\n",global.counter++,global.SnowWinWidth,global.SnowWinHeight,PrevW,PrevH);
    DisplayDimensions();
-   if (prevw != global.SnowWinWidth || prevh != global.SnowWinHeight)
+   if (!global.Trans)
+      HandleX11Cairo();
+   if (PrevW != global.SnowWinWidth || PrevH != global.SnowWinHeight)
    {
-      // if(global.X11cairo) // let op
-      if (!global.Trans)
-      {
-	 P("global.Trans %d\n",global.Trans);
-	 HandleX11Cairo();
-	 //RestartDisplay();
-      }
+      printf(_("Window size changed, now snowing in in %#lx: %s %d+%d %dx%d\n"),global.SnowWin,SnowWinName,global.SnowWinX,global.SnowWinY,global.SnowWinWidth,global.SnowWinHeight);
+      P("Calling RestartDisplay\n");
       RestartDisplay();
-      prevw = global.SnowWinWidth;
-      prevh = global.SnowWinHeight;
+      PrevW = global.SnowWinWidth;
+      PrevH = global.SnowWinHeight;
       SetWindowScale();
    }
+   fflush(stdout);
    return TRUE;
 }
 
@@ -1168,7 +1366,7 @@ int do_draw_all(gpointer widget)
 {
    if (Flags.Done)
       return FALSE;
-   P("do_draw_all %d %p\n",counter++,(void *)widget);
+   P("do_draw_all %d %p\n",global.counter++,(void *)widget);
 
    // this will result in a call off on_draw_event():
    gtk_widget_queue_draw(GTK_WIDGET(widget));
@@ -1179,18 +1377,12 @@ int do_draw_all(gpointer widget)
 // handle callbacks for things whose timings depend on cpufactor
 void HandleCpuFactor()
 {
-   static guint fallen_id=0;
-
    // re-add things whose timing is dependent on cpufactor
    if (Flags.CpuLoad <= 0)
       global.cpufactor = 1;
    else
       global.cpufactor = 100.0/Flags.CpuLoad;
 
-   if (fallen_id)
-      g_source_remove(fallen_id);
-
-   fallen_id = add_to_mainloop(PRIORITY_DEFAULT, time_fallen, do_fallen);
    P("handlecpufactor %f %f %d\n",oldcpufactor,cpufactor,counter++);
    add_to_mainloop(PRIORITY_HIGH, time_init_snow , do_initsnow);  // remove flakes
 
@@ -1214,12 +1406,156 @@ void restart_do_draw_all()
       P("started do_drawit %d %p %f\n",drawit_id, (void *)CairoDC, time_draw_all);
    }
 }
-
+void rectangle_draw(cairo_t *cr)
+{
+   const int lw = 8;
+   cairo_save(cr);
+   cairo_set_source_rgba(cr,1,1,0,0.5);
+   cairo_rectangle(cr,lw/2,lw/2,global.SnowWinWidth-lw,global.SnowWinHeight-lw); 
+   cairo_set_line_width(cr,lw);
+   cairo_stroke(cr);
+   cairo_restore(cr);
+}
 
 int do_stopafter(void *d)
 {
    Flags.Done = 1;
-   printf("Halting because of flag -stopafter\n");
+   printf(_("Halting because of flag %s\n"),"-stopafter");
    return FALSE;
    (void)d;
 }
+
+void mybindtestdomain()
+{
+   global.Language = guess_language();
+#ifdef HAVE_GETTEXT
+   // One would assume that gettext() uses the environment variable LOCPATH
+   // to find locales, but no, it does not.
+   // So, we scan LOCPATH for paths, use them and see if gettext gets a translation
+   // for a text that is known to exist in the ui and should be translated to
+   // something different.
+
+   char *startlocale;
+   (void) startlocale;
+   char *r = setlocale(LC_ALL,"");
+   if (r)
+      startlocale = strdup(r); 
+   else
+      startlocale = NULL;
+   P("startlc: %s\n",startlocale);
+   //if (startlocale == NULL)
+   //  startlocale = strdup("en_US.UTF-8");
+   P("startlc: %s\n",startlocale);
+   P("LC_ALL: %s\n",getenv("LC_ALL"));
+   P("LANG: %s\n",getenv("LANG"));
+   P("LANGUAGE: %s\n",getenv("LANGUAGE"));
+   textdomain (TEXTDOMAIN);
+   char *textd = bindtextdomain(TEXTDOMAIN,LOCALEDIR);
+   (void) textd;
+   P("textdomain:%s\n",textd);
+
+   char *locpath = getenv("LOCPATH");
+   P("LOCPATH: %s\n",locpath);
+   P("LOCALEDIR: %s\n",LOCALEDIR);
+   if (locpath)
+   {
+      char *initial_textdomain = strdup(bindtextdomain(TEXTDOMAIN,NULL));
+      char *mylocpath          = strdup(locpath);
+      int  translation_found   = False;
+      P("Initial textdomain: %s\n",initial_textdomain);
+
+      char *q = mylocpath;
+      while(1)
+      {
+	 char *p = strsep(&q,":");
+	 if (!p)
+	    break;
+	 bindtextdomain(TEXTDOMAIN, p);
+	 //char *trans = _(TESTSTRING);
+	 P("trying '%s' '%s' '%s'\n",p,TESTSTRING,trans);
+	 if(strcmp(_(TESTSTRING),TESTSTRING))
+	 {
+	    // translation found, we are happy
+	    P("Translation: %s\n",trans);
+	    translation_found = True;
+	    break;
+	 }
+      }
+      if (!translation_found)
+      {
+	 P("No translation found in %s\n",locpath);
+	 bindtextdomain(TEXTDOMAIN,initial_textdomain);
+      }
+      free(mylocpath);
+      free(initial_textdomain);
+   }
+
+
+   // todo: naar restart of ui?
+
+   if(0)
+   {
+      if (strcmp(Flags.Language,"sys")) // if "sys" , do not change locale
+      {  // Language != sys
+	 char lc[100];
+	 if (!strcmp(Flags.Language,"en"))
+	    strcpy(lc,"en_US.UTF-8");
+	 else
+	 {
+	    strcpy(lc,Flags.Language);
+	    strcat(lc,"_");
+
+	    int i;
+	    int l = strlen(lc);
+	    for (i=0; i<(int)strlen(Flags.Language); i++)
+	       lc[l+i] = toupper(lc[i]);
+	    lc[l+i] = 0;
+
+	    strcat(lc,".UTF-8");
+	 }
+
+	 // in order for this to work, no gettext call must have been
+	 // made yet:
+
+	 //setenv("LC_ALL",lc,1);
+	 //setenv("LANG",lc,1);
+	 setenv("LANGUAGE",Flags.Language,1);
+	 P("LC_ALL: %s\n",getenv("LC_ALL"));
+	 P("LANG: %s\n",getenv("LANG"));
+	 P("LANGUAGE: %s\n",getenv("LANGUAGE"));
+      }
+      else // Language == sys
+      {
+	 unsetenv("LANGUAGE");
+	 unsetenv("LC_ALL");
+	 if(0)
+	 {
+	    char *l=getenv("LANG");
+	    if(l && !getenv("LANGUAGE"))
+	    {
+	       char *lang = strdup(l);
+	       P("lang: %s\n",lang);
+	       char *p = strchr(lang,'_');
+	       if(p)
+	       {
+		  *p = 0;  // TODO: wrong
+		  P("lang: %s\n",lang);
+		  setenv("LANGUAGE",lang,1);
+	       }
+	       free(lang);
+	    }
+	 }
+	 P("LC_ALL: %s\n",getenv("LC_ALL"));
+	 P("LANG: %s\n",getenv("LANG"));
+	 P("LANGUAGE: %s\n",getenv("LANGUAGE"));
+      }
+      //setenv("LANG","nl_NL.UTF8",1);
+      //setenv("LANGUAGE","nl",1);
+   }
+
+   P("textdomain: %s\n",textdomain(NULL));
+   P("bindtextdomain: %s\n",bindtextdomain(TEXTDOMAIN,NULL));
+   //bind_textdomain_codeset(TEXTDOMAIN,"UTF-8");
+
+#endif
+}
