@@ -2,7 +2,7 @@
 #-# 
 #-# xsnow: let it snow on your desktop
 #-# Copyright (C) 1984,1988,1990,1993-1995,2000-2001 Rick Jansen
-#-# 	      2019,2020,2021 Willem Vermin
+#-# 	      2019,2020,2021,2022,2023,2024 Willem Vermin
 #-# 
 #-# This program is free software: you can redistribute it and/or modify
 #-# it under the terms of the GNU General Public License as published by
@@ -19,11 +19,15 @@
 #-# 
 */
 
+#include <pthread.h>
 #include <stdio.h>
 #include <gtk/gtk.h>
 #include <stdlib.h>
 #include <math.h>
 #include <assert.h>
+
+#include "xsnow-constants.h"
+
 #include "debug.h"
 #include "xsnow.h"
 #include "pixmaps.h"
@@ -48,7 +52,6 @@
 
 #define add_flake_to_mainloop(f) add_to_mainloop1(PRIORITY_HIGH,time_snowflakes,(GSourceFunc)do_UpdateSnowFlake,f)
 
-//static cairo_surface_t **snow_surfaces;
 static float      FlakesPerSecond;
 static int        KillFlakes = 0;  // 1: signal to flakes to kill themselves, and do not generate flakes
 static float      SnowSpeedFactor;
@@ -86,8 +89,8 @@ void snow_init()
    NFlakeTypesVintage = MaxFlakeTypes;
 
    add_random_flakes(EXTRA_FLAKES);   // will change MaxFlakeTypes
-   //                            and create xsnow_xpm, containing
-   //                            vintage and new flakes
+				      //                            and create xsnow_xpm, containing
+				      //                            vintage and new flakes
 
 
    snowPix       = (SnowMap          *)malloc(MaxFlakeTypes*sizeof(SnowMap));
@@ -96,7 +99,6 @@ void snow_init()
 
    for (i=0; i<MaxFlakeTypes; i++)
    {
-      //snow_surfaces[i] = NULL;
       snowPix[i].surface = NULL;
    }
 
@@ -133,10 +135,18 @@ void snow_ui()
    UIDO (NoSnowFlakes, 
 	 if(Flags.NoSnowFlakes)
 	 ClearScreen();                                            );
+   UIDO (UseColor2, 
+	 InitSnowColor();
+	 ClearScreen();
+	);
    UIDO (SnowFlakesFactor               , InitFlakesPerSecond();   );
    UIDOS(SnowColor                      ,
 	 InitSnowColor();
-         InitFallenSnow(); 
+	 //InitFallenSnow(); 
+	 ClearScreen();                                            );
+   UIDOS(SnowColor2                     ,
+	 InitSnowColor();
+	 //InitFallenSnow(); 
 	 ClearScreen();                                            );
    UIDO (SnowSpeedFactor                , InitSnowSpeedFactor();   );
    UIDO (FlakeCountMax                  ,                          );
@@ -166,7 +176,15 @@ void init_snow_pix()
       rp->height = h;
       char **x;
       int lines;
-      xpm_set_color(xsnow_xpm[flake], &x, &lines, Flags.SnowColor);
+
+      if (Flags.UseColor2)
+      {
+	 char color[8];
+	 mixcolors(Flags.SnowColor, Flags.SnowColor2, drand48(), &color[0]);
+	 xpm_set_color(xsnow_xpm[flake], &x, &lines, color);
+      }
+      else
+	 xpm_set_color(xsnow_xpm[flake], &x, &lines, Flags.SnowColor);
       GdkPixbuf *pixbuf = gdk_pixbuf_new_from_xpm_data((const char **)x);
       if (w<1) w=1;
       if (h<1) h=1;
@@ -193,7 +211,6 @@ int snow_draw(cairo_t *cr)
    while( (flake = (Snow *)set_next()) )
    {
       P("snow_draw %d %f\n",counter++,ALPHA);
-      //cairo_set_source_surface (cr, snow_surfaces[flake->whatFlake], flake->rx, flake->ry);
       cairo_set_source_surface (cr, snowPix[flake->whatFlake].surface, flake->rx, flake->ry);
       double alpha = ALPHA;
       if (flake->fluff)
@@ -321,11 +338,18 @@ int do_UpdateSnowFlake(Snow *flake)
    //
    if (!Flags.NoWind)
    {
-      flake->vx += FlakesDT*flake->wsens*(global.NewWind - flake->vx)/flake->m;
+      P("flakevx1: %p %f\n",(void*)flake,flake->vx);
+      float f = FlakesDT*flake->wsens/flake->m;
+      if (f > 0.9)  f =  0.9;
+      if (f < -0.9) f = -0.9;
+      flake->vx += f*(global.NewWind - flake->vx);
       static float speedxmaxes[] = {100.0, 300.0, 600.0,};
-      float speedxmax = speedxmaxes[global.Wind];
-      if(flake->vx > speedxmax) flake->vx = speedxmax;
+      float speedxmax = 2*speedxmaxes[global.Wind];
+      if(fabs(flake->vx)>speedxmax)
+	 P("flakevx:  %p %f %f %f %f %f %f\n",(void*)flake,flake->vx,FlakesDT,flake->wsens,global.NewWind,flake->vx,flake->m);
+      if(flake->vx > speedxmax)  flake->vx =  speedxmax;
       if(flake->vx < -speedxmax) flake->vx = -speedxmax;
+      P("vxa: %f\n",flake->vx);
    }
 
    flake->vy += INITIALYSPEED * (drand48()-0.4)*0.1 ;
@@ -363,6 +387,7 @@ int do_UpdateSnowFlake(Snow *flake)
 
    if (!flake->fluff)
    {
+      Lock_fallen();
       // determine if non-fluffy-flake touches the fallen snow,
       // if so: make the flake inactive.
       // the bottom pixels of the snowflake are at y = NewY + (height of flake)
@@ -374,7 +399,8 @@ int do_UpdateSnowFlake(Snow *flake)
       while(fsnow && !found)
       {
 	 if(!fsnow->win.hidden)
-	    if(fsnow->win.id == 0 ||(fsnow->win.ws == global.CWorkSpace || fsnow->win.sticky))
+	    //if(fsnow->win.id == 0 ||(fsnow->win.ws == global.CWorkSpace || fsnow->win.sticky))
+	    if(fsnow->win.id == 0 ||(IsVisibleFallen(fsnow) || fsnow->win.sticky))
 	    {
 	       if (nx >= fsnow->x && nx <= fsnow->x + fsnow->w &&
 		     ny < fsnow->y+2)
@@ -387,7 +413,7 @@ int do_UpdateSnowFlake(Snow *flake)
 		  for (i = istart; i < imax; i++)
 		     if (ny > fsnow->y - fsnow->acth[i] - 1)
 		     {
-			if(fsnow->acth[i] < fsnow->h/2)
+			if(fsnow->acth[i] < fsnow->desh[i])
 			   UpdateFallenSnowPartial(fsnow,nx - fsnow->x, flakew);
 			if(HandleFallenSnow(fsnow))
 			{
@@ -405,10 +431,14 @@ int do_UpdateSnowFlake(Snow *flake)
 			      fluffify(flake,0.1);
 			   }
 			   if (flake->fluff)
+			   {
+			      Unlock_fallen();
 			      return TRUE;
+			   }
 			   else
 			   {
 			      DelFlake(flake);
+			      Unlock_fallen();
 			      return FALSE;
 			   }
 			}
@@ -419,6 +449,7 @@ int do_UpdateSnowFlake(Snow *flake)
 	    }
 	 fsnow = fsnow->next;
       }
+      Unlock_fallen();
    }
 
    int x  = lrintf(flake->rx);
@@ -469,20 +500,16 @@ int do_UpdateSnowFlake(Snow *flake)
 	    if(found) break;
 	    int ybot = y+flakeh;
 	    int xbot = x+i;
-	    //int in = XRectInRegion(global.TreeRegion,xbot,ybot,1,1);
 	    cairo_rectangle_int_t grect = {xbot,ybot,1,1};
 	    cairo_region_overlap_t in = cairo_region_contains_rectangle(global.TreeRegion,&grect);
-	    //if (in != RectangleIn) // if bottom pixel not in TreeRegion, skip
 	    if (in != CAIRO_REGION_OVERLAP_IN) // if bottom pixel not in TreeRegion, skip
 	       continue;
 	    // move upwards, until pixel is not in TreeRegion
 	    int j;
 	    for (j=ybot-1; j >= y; j--)
 	    {
-	       //int in = XRectInRegion(global.TreeRegion,xbot,j,1,1); 
 	       cairo_rectangle_int_t grect = {xbot,j,1,1};
 	       cairo_region_overlap_t in = cairo_region_contains_rectangle(global.TreeRegion,&grect); 
-	       //if (in != RectangleIn)
 	       if (in != CAIRO_REGION_OVERLAP_IN)
 	       {
 		  // pixel (xbot,j) is snow-on-tree
@@ -558,6 +585,7 @@ void EraseSnowFlake1(Snow *flake)
    P("Erasesnowflake1\n");
    if(global.IsDouble)
       return;
+   P("Erasesnowflake++++++++\n");
    int x = flake->ix-1;
    int y = flake->iy-1;
    int flakew = snowPix[flake->whatFlake].width+2;
@@ -866,6 +894,6 @@ int do_SwitchFlakes(void *d)
 
 void printflake(Snow *flake)
 {
-   printf("flake: %p rx: %6.0f ry: %6.0f fluff: %d freeze: %d ftr: %8.3f ft: %8.3f\n",
-	 (void *)flake,flake->rx,flake->ry,flake->fluff,flake->freeze,flake->flufftimer,flake->flufftime);
+   printf("flake: %p rx: %6.0f ry: %6.0f vx: %6.0f vy: %6.0f ws: %6.0f fluff: %d freeze: %d ftr: %8.3f ft: %8.3f\n",
+	 (void *)flake,flake->rx,flake->ry,flake->vx,flake->vy,flake->wsens,flake->fluff,flake->freeze,flake->flufftimer,flake->flufftime);
 }
